@@ -10,12 +10,13 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
-from tests import unittest, mock, BaseSessionTest, create_session
+from tests import unittest, mock, BaseSessionTest, create_session, ClientHTTPStubber
 from nose.tools import assert_equal
 
 import botocore.session
 from botocore.config import Config
-from botocore.exceptions import ParamValidationError
+from botocore.compat import urlsplit
+from botocore.exceptions import ParamValidationError, ClientError
 from botocore import UNSIGNED
 
 
@@ -34,30 +35,22 @@ class BaseS3OperationTest(BaseSessionTest):
         self.region = 'us-west-2'
         self.client = self.session.create_client(
             's3', self.region)
-        self.session_send_patch = mock.patch('botocore.endpoint.Session.send')
-        self.http_session_send_mock = self.session_send_patch.start()
-
-    def tearDown(self):
-        super(BaseSessionTest, self).tearDown()
-        self.session_send_patch.stop()
+        self.http_stubber = ClientHTTPStubber(self.client)
 
 
 class TestOnlyAsciiCharsAllowed(BaseS3OperationTest):
     def test_validates_non_ascii_chars_trigger_validation_error(self):
-        self.http_session_send_mock.return_value = mock.Mock(status_code=200,
-                                                             headers={},
-                                                             content=b'')
-        with self.assertRaises(ParamValidationError):
-            self.client.put_object(
-                Bucket='foo', Key='bar', Metadata={
-                    'goodkey': 'good', 'non-ascii': u'\u2713'})
+        self.http_stubber.add_response()
+        with self.http_stubber:
+            with self.assertRaises(ParamValidationError):
+                self.client.put_object(
+                    Bucket='foo', Key='bar', Metadata={
+                        'goodkey': 'good', 'non-ascii': u'\u2713'})
 
 
 class TestS3GetBucketLifecycle(BaseS3OperationTest):
     def test_multiple_transitions_returns_one(self):
-        http_response = mock.Mock()
-        http_response.status_code = 200
-        http_response.content = (
+        response_body = (
             '<?xml version="1.0" ?>'
             '<LifecycleConfiguration xmlns="http://s3.amazonaws.'
             'com/doc/2006-03-01/">'
@@ -89,10 +82,10 @@ class TestS3GetBucketLifecycle(BaseS3OperationTest):
             '	</Rule>'
             '</LifecycleConfiguration>'
         ).encode('utf-8')
-        http_response.headers = {}
-        self.http_session_send_mock.return_value = http_response
         s3 = self.session.create_client('s3')
-        response = s3.get_bucket_lifecycle(Bucket='mybucket')
+        with ClientHTTPStubber(s3) as http_stubber:
+            http_stubber.add_response(body=response_body)
+            response = s3.get_bucket_lifecycle(Bucket='mybucket')
         # Each Transition member should have at least one of the
         # transitions provided.
         self.assertEqual(
@@ -129,24 +122,15 @@ class TestS3PutObject(BaseS3OperationTest):
             'Content-Length: 0\r\n'
             'Server: AmazonS3\r\n'
         ).encode('utf-8')
-        http_500_response = mock.Mock()
-        http_500_response.status_code = 500
-        http_500_response.content = non_xml_content
-        http_500_response.headers = {}
-
-        success_response = mock.Mock()
-        success_response.status_code = 200
-        success_response.content = b''
-        success_response.headers = {}
-
-        self.http_session_send_mock.side_effect = [
-            http_500_response, success_response
-        ]
         s3 = self.session.create_client('s3')
-        response = s3.put_object(Bucket='mybucket', Key='mykey', Body=b'foo')
-        # The first response should have been retried even though the xml is
-        # invalid and eventually return the 200 response.
-        self.assertEqual(response['ResponseMetadata']['HTTPStatusCode'], 200)
+        with ClientHTTPStubber(s3) as http_stubber:
+            http_stubber.add_response(status=500, body=non_xml_content)
+            http_stubber.add_response()
+            response = s3.put_object(Bucket='mybucket', Key='mykey', Body=b'foo')
+            # The first response should have been retried even though the xml is
+            # invalid and eventually return the 200 response.
+            self.assertEqual(response['ResponseMetadata']['HTTPStatusCode'], 200)
+            self.assertEqual(len(http_stubber.requests), 2)
 
 
 class TestS3SigV4(BaseS3OperationTest):
@@ -154,17 +138,15 @@ class TestS3SigV4(BaseS3OperationTest):
         super(TestS3SigV4, self).setUp()
         self.client = self.session.create_client(
             's3', self.region, config=Config(signature_version='s3v4'))
-        self.response_mock = mock.Mock()
-        self.response_mock.content = b''
-        self.response_mock.headers = {}
-        self.response_mock.status_code = 200
-        self.http_session_send_mock.return_value = self.response_mock
+        self.http_stubber = ClientHTTPStubber(self.client)
+        self.http_stubber.add_response()
 
     def get_sent_headers(self):
-        return self.http_session_send_mock.mock_calls[0][1][0].headers
+        return self.http_stubber.requests[0].headers
 
     def test_content_md5_set(self):
-        self.client.put_object(Bucket='foo', Key='bar', Body='baz')
+        with self.http_stubber:
+            self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         self.assertIn('content-md5', self.get_sent_headers())
 
     def test_content_sha256_set_if_config_value_is_true(self):
@@ -173,7 +155,10 @@ class TestS3SigV4(BaseS3OperationTest):
         })
         self.client = self.session.create_client(
             's3', self.region, config=config)
-        self.client.put_object(Bucket='foo', Key='bar', Body='baz')
+        self.http_stubber = ClientHTTPStubber(self.client)
+        self.http_stubber.add_response()
+        with self.http_stubber:
+            self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         sent_headers = self.get_sent_headers()
         sha_header = sent_headers.get('x-amz-content-sha256')
         self.assertNotEqual(sha_header, b'UNSIGNED-PAYLOAD')
@@ -184,7 +169,10 @@ class TestS3SigV4(BaseS3OperationTest):
         })
         self.client = self.session.create_client(
             's3', self.region, config=config)
-        self.client.put_object(Bucket='foo', Key='bar', Body='baz')
+        self.http_stubber = ClientHTTPStubber(self.client)
+        self.http_stubber.add_response()
+        with self.http_stubber:
+            self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         sent_headers = self.get_sent_headers()
         sha_header = sent_headers.get('x-amz-content-sha256')
         self.assertEqual(sha_header, b'UNSIGNED-PAYLOAD')
@@ -192,7 +180,8 @@ class TestS3SigV4(BaseS3OperationTest):
     def test_content_sha256_set_if_md5_is_unavailable(self):
         with mock.patch('botocore.auth.MD5_AVAILABLE', False):
             with mock.patch('botocore.handlers.MD5_AVAILABLE', False):
-                self.client.put_object(Bucket='foo', Key='bar', Body='baz')
+                with self.http_stubber:
+                    self.client.put_object(Bucket='foo', Key='bar', Body='baz')
         sent_headers = self.get_sent_headers()
         unsigned = 'UNSIGNED-PAYLOAD'
         self.assertNotEqual(sent_headers['x-amz-content-sha256'], unsigned)
@@ -205,17 +194,15 @@ class TestCanSendIntegerHeaders(BaseSessionTest):
     def test_int_values_with_sigv4(self):
         s3 = self.session.create_client(
             's3', config=Config(signature_version='s3v4'))
-        with mock.patch('botocore.endpoint.Session.send') as mock_send:
-            mock_send.return_value = mock.Mock(status_code=200,
-                                               content=b'',
-                                               headers={})
+        with ClientHTTPStubber(s3) as http_stubber:
+            http_stubber.add_response()
             s3.upload_part(Bucket='foo', Key='bar', Body=b'foo',
                            UploadId='bar', PartNumber=1, ContentLength=3)
-            headers = mock_send.call_args[0][0].headers
+            headers = http_stubber.requests[0].headers
             # Verify that the request integer value of 3 has been converted to
             # string '3'.  This also means we've made it pass the signer which
             # expects string values in order to sign properly.
-            self.assertEqual(headers['Content-Length'], '3')
+            self.assertEqual(headers['Content-Length'], b'3')
 
 
 
@@ -223,77 +210,155 @@ class TestRegionRedirect(BaseS3OperationTest):
     def setUp(self):
         super(TestRegionRedirect, self).setUp()
         self.client = self.session.create_client(
-            's3', 'us-west-2', config=Config(signature_version='s3v4'))
+            's3', 'us-west-2', config=Config(
+                signature_version='s3v4',
+                s3={'addressing_style': 'path'},
+            ))
+        self.http_stubber = ClientHTTPStubber(self.client)
 
-        self.redirect_response = mock.Mock()
-        self.redirect_response.headers = {
-            'x-amz-bucket-region': 'eu-central-1'
+        self.redirect_response = {
+            'status': 301,
+            'headers': {'x-amz-bucket-region': 'eu-central-1'},
+            'body': (
+                b'<?xml version="1.0" encoding="UTF-8"?>\n'
+                b'<Error>'
+                b'    <Code>PermanentRedirect</Code>'
+                b'    <Message>The bucket you are attempting to access must be'
+                b'        addressed using the specified endpoint. Please send '
+                b'        all future requests to this endpoint.'
+                b'    </Message>'
+                b'    <Bucket>foo</Bucket>'
+                b'    <Endpoint>foo.s3.eu-central-1.amazonaws.com</Endpoint>'
+                b'</Error>'
+            )
+        }
+        self.bad_signing_region_response = {
+            'status': 400,
+            'headers': {'x-amz-bucket-region': 'eu-central-1'},
+            'body': (
+                b'<?xml version="1.0" encoding="UTF-8"?>'
+                b'<Error>'
+                b'  <Code>AuthorizationHeaderMalformed</Code>'
+                b'  <Message>the region us-west-2 is wrong; '
+                b'expecting eu-central-1</Message>'
+                b'  <Region>eu-central-1</Region>'
+                b'  <RequestId>BD9AA1730D454E39</RequestId>'
+                b'  <HostId></HostId>'
+                b'</Error>'
+            )
+        }
+        self.success_response = {
+            'status': 200,
+            'headers': {},
+            'body': (
+                b'<?xml version="1.0" encoding="UTF-8"?>\n'
+                b'<ListBucketResult>'
+                b'    <Name>foo</Name>'
+                b'    <Prefix></Prefix>'
+                b'    <Marker></Marker>'
+                b'    <MaxKeys>1000</MaxKeys>'
+                b'    <EncodingType>url</EncodingType>'
+                b'    <IsTruncated>false</IsTruncated>'
+                b'</ListBucketResult>'
+            )
         }
-        self.redirect_response.status_code = 301
-        self.redirect_response.content = (
-            b'<?xml version="1.0" encoding="UTF-8"?>\n'
-            b'<Error>'
-            b'    <Code>PermanentRedirect</Code>'
-            b'    <Message>The bucket you are attempting to access must be '
-            b'        addressed using the specified endpoint. Please send all '
-            b'        future requests to this endpoint.'
-            b'    </Message>'
-            b'    <Bucket>foo</Bucket>'
-            b'    <Endpoint>foo.s3.eu-central-1.amazonaws.com</Endpoint>'
-            b'</Error>')
-
-        self.success_response = mock.Mock()
-        self.success_response.headers = {}
-        self.success_response.status_code = 200
-        self.success_response.content = (
-            b'<?xml version="1.0" encoding="UTF-8"?>\n'
-            b'<ListBucketResult>'
-            b'    <Name>foo</Name>'
-            b'    <Prefix></Prefix>'
-            b'    <Marker></Marker>'
-            b'    <MaxKeys>1000</MaxKeys>'
-            b'    <EncodingType>url</EncodingType>'
-            b'    <IsTruncated>false</IsTruncated>'
-            b'</ListBucketResult>')
 
     def test_region_redirect(self):
-        self.http_session_send_mock.side_effect = [
-            self.redirect_response, self.success_response]
-        response = self.client.list_objects(Bucket='foo')
+        self.http_stubber.add_response(**self.redirect_response)
+        self.http_stubber.add_response(**self.success_response)
+        with self.http_stubber:
+            response = self.client.list_objects(Bucket='foo')
         self.assertEqual(response['ResponseMetadata']['HTTPStatusCode'], 200)
-        self.assertEqual(self.http_session_send_mock.call_count, 2)
+        self.assertEqual(len(self.http_stubber.requests), 2)
 
-        calls = [c[0][0] for c in self.http_session_send_mock.call_args_list]
         initial_url = ('https://s3.us-west-2.amazonaws.com/foo'
                        '?encoding-type=url')
-        self.assertEqual(calls[0].url, initial_url)
+        self.assertEqual(self.http_stubber.requests[0].url, initial_url)
 
         fixed_url = ('https://s3.eu-central-1.amazonaws.com/foo'
                      '?encoding-type=url')
-        self.assertEqual(calls[1].url, fixed_url)
+        self.assertEqual(self.http_stubber.requests[1].url, fixed_url)
 
     def test_region_redirect_cache(self):
-        self.http_session_send_mock.side_effect = [
-            self.redirect_response, self.success_response,
-            self.success_response]
+        self.http_stubber.add_response(**self.redirect_response)
+        self.http_stubber.add_response(**self.success_response)
+        self.http_stubber.add_response(**self.success_response)
+
+        with self.http_stubber:
+            first_response = self.client.list_objects(Bucket='foo')
+            second_response = self.client.list_objects(Bucket='foo')
 
-        first_response = self.client.list_objects(Bucket='foo')
         self.assertEqual(
             first_response['ResponseMetadata']['HTTPStatusCode'], 200)
-        second_response = self.client.list_objects(Bucket='foo')
         self.assertEqual(
             second_response['ResponseMetadata']['HTTPStatusCode'], 200)
 
-        self.assertEqual(self.http_session_send_mock.call_count, 3)
-        calls = [c[0][0] for c in self.http_session_send_mock.call_args_list]
+        self.assertEqual(len(self.http_stubber.requests), 3)
         initial_url = ('https://s3.us-west-2.amazonaws.com/foo'
                        '?encoding-type=url')
-        self.assertEqual(calls[0].url, initial_url)
+        self.assertEqual(self.http_stubber.requests[0].url, initial_url)
 
         fixed_url = ('https://s3.eu-central-1.amazonaws.com/foo'
                      '?encoding-type=url')
-        self.assertEqual(calls[1].url, fixed_url)
-        self.assertEqual(calls[2].url, fixed_url)
+        self.assertEqual(self.http_stubber.requests[1].url, fixed_url)
+        self.assertEqual(self.http_stubber.requests[2].url, fixed_url)
+
+    def test_resign_request_with_region_when_needed(self):
+
+        # Create a client with no explicit configuration so we can
+        # verify the default behavior.
+        client = self.session.create_client('s3', 'us-west-2')
+        with ClientHTTPStubber(client) as http_stubber:
+            http_stubber.add_response(**self.bad_signing_region_response)
+            http_stubber.add_response(**self.success_response)
+            first_response = client.list_objects(Bucket='foo')
+            self.assertEqual(
+                first_response['ResponseMetadata']['HTTPStatusCode'], 200)
+
+            self.assertEqual(len(http_stubber.requests), 2)
+            initial_url = ('https://foo.s3.us-west-2.amazonaws.com/'
+                           '?encoding-type=url')
+            self.assertEqual(http_stubber.requests[0].url, initial_url)
+
+            fixed_url = ('https://foo.s3.eu-central-1.amazonaws.com/'
+                         '?encoding-type=url')
+            self.assertEqual(http_stubber.requests[1].url, fixed_url)
+
+    def test_resign_request_in_us_east_1(self):
+        region_headers = {'x-amz-bucket-region': 'eu-central-1'}
+
+        # Verify that the default behavior in us-east-1 will redirect
+        client = self.session.create_client('s3', 'us-east-1')
+        with ClientHTTPStubber(client) as http_stubber:
+            http_stubber.add_response(status=400)
+            http_stubber.add_response(status=400, headers=region_headers)
+            http_stubber.add_response(headers=region_headers)
+            http_stubber.add_response()
+            response = client.head_object(Bucket='foo', Key='bar')
+            self.assertEqual(response['ResponseMetadata']['HTTPStatusCode'], 200)
+
+            self.assertEqual(len(http_stubber.requests), 4)
+            initial_url = ('https://foo.s3.amazonaws.com/bar')
+            self.assertEqual(http_stubber.requests[0].url, initial_url)
+
+            fixed_url = ('https://foo.s3.eu-central-1.amazonaws.com/bar')
+            self.assertEqual(http_stubber.requests[-1].url, fixed_url)
+
+    def test_resign_request_in_us_east_1_fails(self):
+        region_headers = {'x-amz-bucket-region': 'eu-central-1'}
+
+        # Verify that the final 400 response is propagated
+        # back to the user.
+        client = self.session.create_client('s3', 'us-east-1')
+        with ClientHTTPStubber(client) as http_stubber:
+            http_stubber.add_response(status=400)
+            http_stubber.add_response(status=400, headers=region_headers)
+            http_stubber.add_response(headers=region_headers)
+            # The final request still fails with a 400.
+            http_stubber.add_response(status=400)
+            with self.assertRaises(ClientError) as e:
+                client.head_object(Bucket='foo', Key='bar')
+            self.assertEqual(len(http_stubber.requests), 4)
 
 
 class TestGeneratePresigned(BaseS3OperationTest):
@@ -403,42 +468,45 @@ def test_correct_url_used_for_s3():
     # The default behavior for sigv2. DNS compatible buckets
     yield t.case(region='us-west-2', bucket='bucket', key='key',
                  signature_version='s3',
-                 expected_url='https://bucket.s3.amazonaws.com/key')
+                 expected_url='https://bucket.s3.us-west-2.amazonaws.com/key')
     yield t.case(region='us-east-1', bucket='bucket', key='key',
                  signature_version='s3',
                  expected_url='https://bucket.s3.amazonaws.com/key')
     yield t.case(region='us-west-1', bucket='bucket', key='key',
                  signature_version='s3',
-                 expected_url='https://bucket.s3.amazonaws.com/key')
+                 expected_url='https://bucket.s3.us-west-1.amazonaws.com/key')
     yield t.case(region='us-west-1', bucket='bucket', key='key',
                  signature_version='s3', is_secure=False,
-                 expected_url='http://bucket.s3.amazonaws.com/key')
+                 expected_url='http://bucket.s3.us-west-1.amazonaws.com/key')
 
-    # The default behavior for sigv4. DNS compatible buckets still get path
-    # style addresses.
+    # Virtual host addressing is independent of signature version.
     yield t.case(region='us-west-2', bucket='bucket', key='key',
                  signature_version='s3v4',
                  expected_url=(
-                     'https://s3.us-west-2.amazonaws.com/bucket/key'))
+                     'https://bucket.s3.us-west-2.amazonaws.com/key'))
     yield t.case(region='us-east-1', bucket='bucket', key='key',
                  signature_version='s3v4',
-                 expected_url='https://s3.amazonaws.com/bucket/key')
+                 expected_url='https://bucket.s3.amazonaws.com/key')
     yield t.case(region='us-west-1', bucket='bucket', key='key',
                  signature_version='s3v4',
                  expected_url=(
-                     'https://s3.us-west-1.amazonaws.com/bucket/key'))
+                     'https://bucket.s3.us-west-1.amazonaws.com/key'))
     yield t.case(region='us-west-1', bucket='bucket', key='key',
                  signature_version='s3v4', is_secure=False,
                  expected_url=(
-                     'http://s3.us-west-1.amazonaws.com/bucket/key'))
+                     'http://bucket.s3.us-west-1.amazonaws.com/key'))
+    yield t.case(
+        region='us-west-1', bucket='bucket-with-num-1', key='key',
+        signature_version='s3v4', is_secure=False,
+        expected_url='http://bucket-with-num-1.s3.us-west-1.amazonaws.com/key')
 
     # Regions outside of the 'aws' partition.
-    # We're expecting path style because this is the default with
-    # 's3v4'.
+    # These should still default to virtual hosted addressing
+    # unless explicitly configured otherwise.
     yield t.case(region='cn-north-1', bucket='bucket', key='key',
                  signature_version='s3v4',
                  expected_url=(
-                     'https://s3.cn-north-1.amazonaws.com.cn/bucket/key'))
+                     'https://bucket.s3.cn-north-1.amazonaws.com.cn/key'))
     # This isn't actually supported because cn-north-1 is sigv4 only,
     # but we'll still double check that our internal logic is correct
     # when building the expected url.
@@ -479,6 +547,25 @@ def test_correct_url_used_for_s3():
     yield t.case(
         region='us-east-1', bucket='bucket.dot', key='key',
         expected_url='https://s3.amazonaws.com/bucket.dot/key')
+    yield t.case(
+        region='us-east-1', bucket='BucketName', key='key',
+        expected_url='https://s3.amazonaws.com/BucketName/key')
+    yield t.case(
+        region='us-west-1', bucket='bucket_name', key='key',
+        expected_url='https://s3.us-west-1.amazonaws.com/bucket_name/key')
+    yield t.case(
+        region='us-west-1', bucket='-bucket-name', key='key',
+        expected_url='https://s3.us-west-1.amazonaws.com/-bucket-name/key')
+    yield t.case(
+        region='us-west-1', bucket='bucket-name-', key='key',
+        expected_url='https://s3.us-west-1.amazonaws.com/bucket-name-/key')
+    yield t.case(
+        region='us-west-1', bucket='aa', key='key',
+        expected_url='https://s3.us-west-1.amazonaws.com/aa/key')
+    yield t.case(
+        region='us-west-1', bucket='a'*64, key='key',
+        expected_url=('https://s3.us-west-1.amazonaws.com/%s/key' % ('a' * 64))
+    )
 
     # Custom endpoint url should always be used.
     yield t.case(
@@ -522,15 +609,14 @@ def test_correct_url_used_for_s3():
         s3_config=virtual_hosting,
         expected_url='https://bucket.s3.us-gov-west-1.amazonaws.com/key')
 
-    # Test restricted regions not do virtual host by default
     yield t.case(
         region='us-gov-west-1', bucket='bucket', key='key',
         signature_version='s3',
-        expected_url='https://s3.us-gov-west-1.amazonaws.com/bucket/key')
+        expected_url='https://bucket.s3.us-gov-west-1.amazonaws.com/key')
     yield t.case(
         region='fips-us-gov-west-1', bucket='bucket', key='key',
         signature_version='s3',
-        expected_url='https://s3-fips-us-gov-west-1.amazonaws.com/bucket/key')
+        expected_url='https://bucket.s3-fips-us-gov-west-1.amazonaws.com/key')
 
 
     # Test path style addressing.
@@ -608,11 +694,11 @@ def test_correct_url_used_for_s3():
     yield t.case(
         region='us-east-1', bucket='bucket', key='key',
         s3_config=use_dualstack, signature_version='s3v4',
-        expected_url='https://s3.dualstack.us-east-1.amazonaws.com/bucket/key')
+        expected_url='https://bucket.s3.dualstack.us-east-1.amazonaws.com/key')
     yield t.case(
         region='us-west-2', bucket='bucket', key='key',
         s3_config=use_dualstack, signature_version='s3v4',
-        expected_url='https://s3.dualstack.us-west-2.amazonaws.com/bucket/key')
+        expected_url='https://bucket.s3.dualstack.us-west-2.amazonaws.com/key')
     # Non DNS compatible buckets use path style for dual stack.
     yield t.case(
         region='us-west-2', bucket='bucket.dot', key='key',
@@ -725,15 +811,12 @@ def _verify_expected_endpoint_url(region, bucket, key, s3_config,
                                   is_secure=True,
                                   customer_provided_endpoint=None,
                                   expected_url=None, signature_version=None):
-    http_response = mock.Mock()
-    http_response.status_code = 200
-    http_response.headers = {}
-    http_response.content = b''
     environ = {}
     with mock.patch('os.environ', environ):
         environ['AWS_ACCESS_KEY_ID'] = 'access_key'
         environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
         environ['AWS_CONFIG_FILE'] = 'no-exist-foo'
+        environ['AWS_SHARED_CREDENTIALS_FILE'] = 'no-exist-foo'
         session = create_session()
         session.config_filename = 'no-exist-foo'
         config = Config(
@@ -743,9 +826,134 @@ def _verify_expected_endpoint_url(region, bucket, key, s3_config,
         s3 = session.create_client('s3', region_name=region, use_ssl=is_secure,
                                    config=config,
                                    endpoint_url=customer_provided_endpoint)
-        with mock.patch('botocore.endpoint.Session.send') as mock_send:
-            mock_send.return_value = http_response
-            s3.put_object(Bucket=bucket,
-                          Key=key, Body=b'bar')
-            request_sent = mock_send.call_args[0][0]
-            assert_equal(request_sent.url, expected_url)
+        with ClientHTTPStubber(s3) as http_stubber:
+            http_stubber.add_response()
+            s3.put_object(Bucket=bucket, Key=key, Body=b'bar')
+            assert_equal(http_stubber.requests[0].url, expected_url)
+
+
+def _create_s3_client(region, is_secure, endpoint_url, s3_config,
+                      signature_version):
+    environ = {}
+    with mock.patch('os.environ', environ):
+        environ['AWS_ACCESS_KEY_ID'] = 'access_key'
+        environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
+        environ['AWS_CONFIG_FILE'] = 'no-exist-foo'
+        environ['AWS_SHARED_CREDENTIALS_FILE'] = 'no-exist-foo'
+        session = create_session()
+        session.config_filename = 'no-exist-foo'
+        config = Config(
+            signature_version=signature_version,
+            s3=s3_config
+        )
+        s3 = session.create_client('s3', region_name=region, use_ssl=is_secure,
+                                   config=config,
+                                   endpoint_url=endpoint_url)
+        return s3
+
+
+def test_addressing_for_presigned_urls():
+    # See TestGeneratePresigned for more detailed test cases
+    # on presigned URLs.  Here's we're just focusing on the
+    # adddressing mode used for presigned URLs.
+    # We special case presigned URLs due to backwards
+    # compatibility.
+    t = S3AddressingCases(_verify_presigned_url_addressing)
+
+    # us-east-1, or the "global" endpoint. A signature version of
+    # None means the user doesn't have signature version configured.
+    yield t.case(region='us-east-1', bucket='bucket', key='key',
+                 signature_version=None,
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-1', bucket='bucket', key='key',
+                 signature_version='s3',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-1', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-1', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 s3_config={'addressing_style': 'path'},
+                 expected_url='https://s3.amazonaws.com/bucket/key')
+
+    # A region that supports both 's3' and 's3v4'.
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version=None,
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version='s3',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 s3_config={'addressing_style': 'path'},
+                 expected_url='https://s3.us-west-2.amazonaws.com/bucket/key')
+
+    # An 's3v4' only region.
+    yield t.case(region='us-east-2', bucket='bucket', key='key',
+                 signature_version=None,
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-2', bucket='bucket', key='key',
+                 signature_version='s3',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-2', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+    yield t.case(region='us-east-2', bucket='bucket', key='key',
+                 signature_version='s3v4',
+                 s3_config={'addressing_style': 'path'},
+                 expected_url='https://s3.us-east-2.amazonaws.com/bucket/key')
+
+    # Dualstack endpoints
+    yield t.case(
+        region='us-west-2', bucket='bucket', key='key',
+        signature_version=None,
+        s3_config={'use_dualstack_endpoint': True},
+        expected_url='https://bucket.s3.dualstack.us-west-2.amazonaws.com/key')
+    yield t.case(
+        region='us-west-2', bucket='bucket', key='key',
+        signature_version='s3',
+        s3_config={'use_dualstack_endpoint': True},
+        expected_url='https://bucket.s3.dualstack.us-west-2.amazonaws.com/key')
+    yield t.case(
+        region='us-west-2', bucket='bucket', key='key',
+        signature_version='s3v4',
+        s3_config={'use_dualstack_endpoint': True},
+        expected_url='https://bucket.s3.dualstack.us-west-2.amazonaws.com/key')
+
+    # Accelerate
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version=None,
+                 s3_config={'use_accelerate_endpoint': True},
+                 expected_url='https://bucket.s3-accelerate.amazonaws.com/key')
+
+    # A region that we don't know about.
+    yield t.case(region='us-west-50', bucket='bucket', key='key',
+                 signature_version=None,
+                 expected_url='https://bucket.s3.amazonaws.com/key')
+
+    # Customer provided URL results in us leaving the host untouched.
+    yield t.case(region='us-west-2', bucket='bucket', key='key',
+                 signature_version=None,
+                 customer_provided_endpoint='https://foo.com/',
+                 expected_url='https://foo.com/bucket/key')
+
+
+def _verify_presigned_url_addressing(region, bucket, key, s3_config,
+                                     is_secure=True,
+                                     customer_provided_endpoint=None,
+                                     expected_url=None,
+                                     signature_version=None):
+    s3 = _create_s3_client(region=region, is_secure=is_secure,
+                           endpoint_url=customer_provided_endpoint,
+                           s3_config=s3_config,
+                           signature_version=signature_version)
+    url = s3.generate_presigned_url(
+        'get_object', {'Bucket': bucket, 'Key': key})
+    # We're not trying to verify the params for URL presigning,
+    # those are tested elsewhere.  We just care about the hostname/path.
+    parts = urlsplit(url)
+    actual = '%s://%s%s' % parts[:3]
+    assert_equal(actual, expected_url)
