@@ -288,11 +288,12 @@ def test_client_can_retry_request_properly():
 def _make_client_call_with_errors(client, operation_name, kwargs):
     operation = getattr(client, xform_name(operation_name))
     exception = ConnectionClosedError(endpoint_url='https://mock.eror')
-    with ClientHTTPStubber(client) as http_stubber:
+    with ClientHTTPStubber(client, strict=False) as http_stubber:
         http_stubber.responses.append(exception)
-        http_stubber.responses.append(None)
         try:
             response = operation(**kwargs)
         except ClientError as e:
             assert False, ('Request was not retried properly, '
                            'received error:\n%s' % pformat(e))
+        # Ensure we used the stubber as we're not using it in strict mode
+        assert len(http_stubber.responses) == 0, 'Stubber was not used!'
