@@ -24,6 +24,7 @@ import botocore.session
 from botocore.compat import OrderedDict
 from botocore.exceptions import ParamValidationError, MD5UnavailableError
 from botocore.exceptions import AliasConflictParameterError
+from botocore.exceptions import MissingServiceIdError
 from botocore.awsrequest import AWSRequest
 from botocore.compat import quote, six
 from botocore.config import Config
@@ -31,8 +32,10 @@ from botocore.docs.bcdoc.restdoc import DocumentStructure
 from botocore.docs.params import RequestParamsDocumenter
 from botocore.docs.example import RequestExampleDocumenter
 from botocore.hooks import HierarchicalEmitter
+from botocore.loaders import Loader
 from botocore.model import OperationModel, ServiceModel, ServiceId
 from botocore.model import DenormalizedStructureBuilder
+from botocore.session import Session
 from botocore.signers import RequestSigner
 from botocore.credentials import Credentials
 from botocore import handlers
@@ -557,6 +560,23 @@ class TestHandlers(BaseSessionTest):
                                               service_name='foo')
         self.assertFalse(session.register.called)
 
+    def test_register_retry_for_handlers_with_no_service_id(self):
+        service_data = {
+            'metadata': {
+                'endpointPrefix': 'foo',
+            },
+        }
+        session = mock.Mock(spec=Session)
+        loader = mock.Mock(spec=Loader)
+        session.get_component.return_value = loader
+        service_name = 'foo'
+        with self.assertRaisesRegexp(MissingServiceIdError, service_name):
+            handlers.register_retries_for_service(
+                service_data=service_data,
+                session=session,
+                service_name=service_name,
+            )
+
     def test_register_retry_handlers(self):
         service_data = {
             'metadata': {
@@ -889,7 +909,7 @@ class TestHandlers(BaseSessionTest):
         }
         handlers.decode_list_object_v2(parsed, context={})
         self.assertEqual(parsed['Contents'][0]['Key'], u'%C3%A7%C3%B6s%25asd')
-        
+
     def test_decode_list_objects_v2_with_delimiter(self):
         parsed = {
             'Delimiter': "%C3%A7%C3%B6s%25%20asd%08+c",
@@ -907,7 +927,7 @@ class TestHandlers(BaseSessionTest):
         context = {'encoding_type_auto_set': True}
         handlers.decode_list_object_v2(parsed, context=context)
         self.assertEqual(parsed['Prefix'], u'\xe7\xf6s% asd\x08 c')
-        
+
     def test_decode_list_objects_v2_does_not_decode_continuationtoken(self):
         parsed = {
             'ContinuationToken': "%C3%A7%C3%B6s%25%20asd%08+c",
@@ -917,7 +937,7 @@ class TestHandlers(BaseSessionTest):
         handlers.decode_list_object_v2(parsed, context=context)
         self.assertEqual(
             parsed['ContinuationToken'], u"%C3%A7%C3%B6s%25%20asd%08+c")
-        
+
     def test_decode_list_objects_v2_with_startafter(self):
         parsed = {
             'StartAfter': "%C3%A7%C3%B6s%25%20asd%08+c",
@@ -926,7 +946,7 @@ class TestHandlers(BaseSessionTest):
         context = {'encoding_type_auto_set': True}
         handlers.decode_list_object_v2(parsed, context=context)
         self.assertEqual(parsed['StartAfter'], u'\xe7\xf6s% asd\x08 c')
-        
+
     def test_decode_list_objects_v2_with_common_prefixes(self):
         parsed = {
             'CommonPrefixes': [{'Prefix': "%C3%A7%C3%B6s%25%20asd%08+c"}],
@@ -936,7 +956,7 @@ class TestHandlers(BaseSessionTest):
         handlers.decode_list_object_v2(parsed, context=context)
         self.assertEqual(parsed['CommonPrefixes'][0]['Prefix'],
                          u'\xe7\xf6s% asd\x08 c')
-        
+
     def test_get_bucket_location_optional(self):
         # This handler should no-op if another hook (i.e. stubber) has already
         # filled in response
@@ -1330,3 +1350,36 @@ class TestCommandAlias(unittest.TestCase):
 
         response = alias(client=client)()
         self.assertEqual(response, 'bar')
+
+
+class TestPrependToHost(unittest.TestCase):
+    def setUp(self):
+        self.hoister = handlers.HeaderToHostHoister('test-header')
+
+    def _prepend_to_host(self, url, prepend_string):
+        params = {
+            'headers': {
+                'test-header': prepend_string,
+            },
+            'url': url,
+        }
+        self.hoister.hoist(params=params)
+        return params['url']
+
+    def test_does_prepend_to_host(self):
+        prepended = self._prepend_to_host('https://bar.example.com/', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.example.com/')
+
+    def test_does_prepend_to_host_with_http(self):
+        prepended = self._prepend_to_host('http://bar.example.com/', 'foo')
+        self.assertEqual(prepended, 'http://foo.bar.example.com/')
+
+    def test_does_prepend_to_host_with_path(self):
+        prepended = self._prepend_to_host(
+            'https://bar.example.com/path', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.example.com/path')
+
+    def test_does_prepend_to_host_with_more_components(self):
+        prepended = self._prepend_to_host(
+            'https://bar.baz.example.com/path', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.baz.example.com/path')
